/**
 * Created by janis dähne.
 */

import * as constants from '../../../constants'
import axios, {AxiosRequestConfig, AxiosResponse, CancelToken, CancelTokenSource} from 'axios'
import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../state/reducers";

import PanelGrid from '../../panelSystem/panelGrid/panelGrid'
import Panel from '../../panelSystem/panel/panel'
import TabContentHost from '../../panelSystem/panel/tabContentHost'

import {RouteComponentProps, withRouter} from "react-router";
import {parseBoolean, parseId} from "../../../helpers/convertersAndTransformers";
import {
  loadExerciseEditor,
  setAddAssetFilesModalFilterText,
  setAddAssetFilesModalUploadPercentage,
  setBotLeftPanelViewSize,
  setBotRightPanelViewSize,
  setBottomLeftSelectedTabIndex,
  setBottomRightSelectedTabIndex,
  setExerciseEditorHDividerPosOffset,
  setExerciseEditorVDividerPosOffset,
  setIsAddAssetFilesModalDisplayed,
  setIsChangeExerciseEditorSettingsDialogDisplayed,
  setTopLeftPanelViewSize,
  setTopLeftSelectedTabIndex,
  setTopRightPanelViewSize,
  setTopRightSelectedTabIndex
} from "../../../state/actions/exerciseEditorSite/exerciseEditorSiteActions";
import Spinner from "../../helpers/spinner";
//---bottom right panel
import ChangeTestsPanelHeaderBar from './exerciseTestsPanel/editorTestsPanelHeaderBar'
import ChangeTestsPanelView from './exerciseTestsPanel/editorTestsPanelView'
//---bottom left panel
import NotePanelView from './notePanel/notePanelView'

import TagsPanelView from './tagsPanel/tagsPanelView'
import TagsPanelHeaderBar from './tagsPanel/tagsPanelHeaderBar'

import AssetsPanelView from './assetsPanel/assetsPanelView'
import AssetsPanelHeaderBar from './assetsPanel/assetsPanelHeaderBar'

import TemplatesPanelView from './codeTemplatesPanel/codeTemplatesPanelView'
import TemplatesPanelHeaderBar from './codeTemplatesPanel/codeTemplatesPanelHeaderBar'
//---top right panel
import PreviewPanelView from './previewPanel/previewPanelView'
import ExerciseSettingsPanelView from './exerciseSettingsPanel/exerciseSettingsPanelView'
import ExerciseSyntaxHelpPanelView from './exerciseSyntaxHelpPanel/exerciseSyntaxHelpPanelView'
import ExerciseSyntaxHelpPanelHeaderBar from './exerciseSyntaxHelpPanel/exerciseSyntaxHelpPanelHeaderBar'
//--- top left panel
import DescriptionEditorPanelView from './descriptionEditorPanel/descriptionEditorPanelView'
import DescriptionEditorHeaderBar from './descriptionEditorPanel/descriptionEditorHeaderBar'
//--- dialogs
import ChangeTestDialog from './exerciseTestsPanel/dialogs/changeTestDialog'
import ChangeSelectedTagsDialog from './tagsPanel/dialogs/changeSelectedTagsDialog'
import ChangeCodeTemplateDialog from './codeTemplatesPanel/dialogs/changeCodeTemplateDialog'
import EditorSettingsDialog from '../../editorSettingsDialog/editorSettingsDialog'


import {LoginTimeoutHelper} from "../../../helpers/loginTimeoutHelper";
import {
  setDescriptionContent,
  setExerciseEditorAssets
} from "../../../state/actions/exerciseEditorSite/subSets/editorExerciseDescriptionActions";
import {createAndRefreshAssetUrlsAndObjects, freeAssetUrlsAndObjects} from "../../../helpers/assetHelper";
import {getI18n} from "../../../../i18n/i18nRoot";
import Logger from '../../../helpers/logger'
import {AceEditorsStorageHelper} from '../../codeEditors/aceEditors/aceEditorsStorageHelper'
import {ErrorHelper} from '../../../helpers/errorHelper'
import ZoomPreviewPanelActions from './previewPanel/zoomPreviewPanelActions'

import AddAssetFilesModal from './addAssetFilesModal/addAssetFilesModal'
import {FilePreviewFromBackend} from '../../../types/exerciseEditor'
import {
  _getSingleExerciseDescriptionAssetFileAsync,
  attachExitingFileToExerciseDescriptionAsync,
  detachFileFromExerciseDescriptionAsync,
  getMarkdownAssetPreviewsAsync, getTestAssetPreviewsAsync
} from '../../../state/actions/exerciseEditorSite/exerciseEditorCrudActions'
import {uploadExerciseDescriptionAssetFile} from '../../../communicationLayer/exerciseEditorLayer'

//const css = require('./styles.styl');

export interface MyProps {
  //readonly test: string
}

const mapStateToProps = (rootState: RootState /*, props: MyProps*/) => {
  return {
    //test0: rootState...
    //test: props.test

    editorExercise: rootState.exerciseEditorState.editorExercise,

    allMarkdownAssetFilePreviews: rootState.exerciseEditorState.allMarkdownAssetFilePreviews,
    isAddAssetFilesModalBusy: rootState.exerciseEditorState.isAddAssetFilesModalBusy,
    addAssetFilesModalFilterText: rootState.exerciseEditorState.addAssetFilesModalFilterText,
    addAssetFilesModalUploadPercentage: rootState.exerciseEditorState.addAssetFilesModalUploadPercentage,

    taskDescriptionFontSizeInPx: rootState.globalUiState.taskDescriptionFontSizeInPx,

    exerciseDescriptionContent: rootState.exerciseEditorState.editorExercise.exerciseDescription.content,

    isLoading: rootState.exerciseEditorState.isLoading
      || rootState.testTypesState.isLoading
      || rootState.pLangsState.isLoading
      || rootState.manageTagsState.isLoading
      || rootState.manageGroupExercisesState.isLoading,

    topRightSelectedTabIndex: rootState.exerciseEditorState.topRightSelectedTabIndex,
    topLeftSelectedTabIndex: rootState.exerciseEditorState.topLeftSelectedTabIndex,
    bottomRightSelectedTabIndex: rootState.exerciseEditorState.bottomRightSelectedTabIndex,
    bottomLeftSelectedTabIndex: rootState.exerciseEditorState.bottomLeftSelectedTabIndex,

    topRightPanelViewSize: rootState.exerciseEditorState.topRightPanelViewSize,
    topLeftPanelViewSize: rootState.exerciseEditorState.topLeftPanelViewSize,
    botRightPanelViewSize: rootState.exerciseEditorState.botRightPanelViewSize,
    botLeftPanelViewSize: rootState.exerciseEditorState.botLeftPanelViewSize,

    isChangeTestDialogDisplayed: rootState.exerciseEditorState.isChangeTestDialogDisplayed,
    isChangeSelectedTagsDialogDisplayed: rootState.exerciseEditorState.isChangeSelectedTagsDialogDisplayed,
    isChangeCodeTemplateDialogDisplayed: rootState.exerciseEditorState.isChangeCodeTemplateDialogDisplayed,
    isChangeEditorSettingsDialogDisplayed: rootState.exerciseEditorState.isChangeEditorSettingsDialogDisplayed,

    isAddAssetFilesModalDisplayed: rootState.exerciseEditorState.isAddAssetFilesModalDisplayed,

    assets: rootState.exerciseEditorState.editorExercise.exerciseDescription.assets,

    isNoDialogDisplayed: !rootState.exerciseEditorState.isChangeTestDialogDisplayed
      && !rootState.exerciseEditorState.isChangeSelectedTagsDialogDisplayed
      && !rootState.exerciseEditorState.isChangeCodeTemplateDialogDisplayed
      && !rootState.exerciseEditorState.isChangeEditorSettingsDialogDisplayed
    ,

    vDividerPosOffset: rootState.exerciseEditorState.vDividerPosOffset,
    hDividerPosOffset: rootState.exerciseEditorState.hDividerPosOffset,
    langId: rootState.i18nState.langId
  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here
                                                                        loadExerciseEditor,

                                                                        setIsChangeExerciseEditorSettingsDialogDisplayed,

                                                                        setTopLeftSelectedTabIndex,
                                                                        setTopRightSelectedTabIndex,
                                                                        setBottomRightSelectedTabIndex,
                                                                        setBottomLeftSelectedTabIndex,

                                                                        setExerciseEditorHDividerPosOffset,
                                                                        setExerciseEditorVDividerPosOffset,

                                                                        setTopRightPanelViewSize,
                                                                        setTopLeftPanelViewSize,
                                                                        setBotRightPanelViewSize,
                                                                        setBotLeftPanelViewSize,


                                                                        setExerciseEditorAssets,
                                                                        setDescriptionContent,

                                                                        setIsAddAssetFilesModalDisplayed,
                                                                        setAddAssetFilesModalFilterText,
                                                                        setAddAssetFilesModalUploadPercentage,
                                                                        attachExitingFileToExerciseDescriptionAsync,
                                                                        detachFileFromExerciseDescriptionAsync,
                                                                        getMarkdownAssetPreviewsAsync,

                                                                        _getSingleExerciseDescriptionAssetFileAsync,

                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;


// tslint:disable-next-line
let currentFileUploadCancelToken: CancelTokenSource | null = null

interface MatchedProps {

  readonly comesFromOwnExercisesSite: string
  /**
   * the exercise id to load or undefined to create a new one
   */
  readonly exerciseId: string | undefined
}

const exerciseEditorTaskPrintId = 'exercise-editor-task-print-id'

class exerciseEditorSite extends React.Component<Props & RouteComponentProps<MatchedProps>, any> {


  async componentWillMount(): Promise<void> {

    const exerciseId = parseId(this.props.match.params.exerciseId)
    // noinspection TsLint
    let comesFromOwnExercisesSite = parseBoolean(this.props.match.params.comesFromOwnExercisesSite)

    if (comesFromOwnExercisesSite === null) {
      Logger.error('could not get comesFromOwnExercisesSite from url, using true')
      comesFromOwnExercisesSite = true
    }

    if (exerciseId !== null) {
      //Load

      if (LoginTimeoutHelper.isReLogin && exerciseId === LoginTimeoutHelper.lastKnownExerciseEditorId) {
        //do not reload or reset site state...
        //user re logged in he/she wants to continue

        createAndRefreshAssetUrlsAndObjects(this.props.editorExercise.exerciseDescription.assets)

        return
      } else {
        LoginTimeoutHelper.lastKnownExerciseEditorId = exerciseId
      }


      try {
        //reset ace storage
        AceEditorsStorageHelper.removeAll()
        await this.props.loadExerciseEditor(exerciseId, false, comesFromOwnExercisesSite)
      } catch (err) {
        ErrorHelper.makeServerErrorDialog(err, this.props.langId, 'Error', 'Could not load site data')
      }

    } else {
      //create

      if (this.props.editorExercise.id <= 0) {
        if (LoginTimeoutHelper.isReLogin && this.props.editorExercise
          && this.props.editorExercise.id === LoginTimeoutHelper.lastKnownExerciseEditorId
        ) {
          //do not reload or reset site state...
          //user re logged in he/she want's to continue
          return
        } else {
          //this would give us an old id because loadExerciseEditor will create a new one...
          //LoginTimeoutHelper.lastKnownExerciseEditorId = this.props.editorExercise.id
        }
      }

      //assume create exercise
      try {
        //reset ace storage
        AceEditorsStorageHelper.removeAll()
        await this.props.loadExerciseEditor(null, false, comesFromOwnExercisesSite)
      } catch (err) {
        ErrorHelper.makeServerErrorDialog(err, this.props.langId, 'Error', 'Could not load site data')
      }

      //grab the new exercise id
      LoginTimeoutHelper.lastKnownExerciseEditorId = this.props.editorExercise.id
    }
  }

  componentWillUnmount(): void {
    //if we not freeing we would leak memory... if we just exit the site
    freeAssetUrlsAndObjects()
  }

  render(): JSX.Element {

    return (
      <div className="fh">

        {
          this.props.isLoading &&
          <Spinner/>
        }

        {
          //we need an old exercise (saved) else we cannot attach files to it!
          <AddAssetFilesModal

            title={getI18n(this.props.langId, "Manage asset files")}
            isDisplayed={this.props.isAddAssetFilesModalDisplayed && this.props.editorExercise.id > 0}
            isBusy={this.props.isAddAssetFilesModalBusy}
            uploadInputId={'exercise-description-upload'}

            onUploadFile={async (file) => {

              this.props.setIsAddAssetFilesModalDisplayed(true, true)
              this.props.setAddAssetFilesModalUploadPercentage(0)

              currentFileUploadCancelToken = axios.CancelToken.source()

              try {
                const assetFilePreview = await uploadExerciseDescriptionAssetFile(this.props.editorExercise.id, file,
                                                                                  currentFileUploadCancelToken.token,
                                                                                  this.props.setAddAssetFilesModalUploadPercentage
                )

                currentFileUploadCancelToken = null

                //reset for download
                //for some reason download shows total of 0, see https://github.com/axios/axios/issues/1591
                this.props.setAddAssetFilesModalUploadPercentage('active')
                Logger.debug('exerciseEditorSite',
                             `upload finished (new id: ${assetFilePreview.id}), now downloading file content again...`
                )

                //this will download the file ... even if we uploaded it (this is ok, we check that everything went well)
                await this.props._getSingleExerciseDescriptionAssetFileAsync(this.props.editorExercise.id,
                                                                             assetFilePreview.id,
                                                                             // this.props.setAddAssetFilesModalUploadPercentage
                )

                //renew all asset preview list (could have changed) this is easier than to manage all special cases
                //e.g. detach file... was it deleted or just detached??
                // this.props.finishAddAssetFilesModalUpload(assetFilePreview)
                await this.props.getMarkdownAssetPreviewsAsync(this.props.editorExercise.id)

              } catch (err) {
                Logger.error('could not upload file, err: ' + err.message)
              } finally {
                currentFileUploadCancelToken = null
                this.props.setIsAddAssetFilesModalDisplayed(true, false)
                this.props.setAddAssetFilesModalUploadPercentage(null)
              }

            }}

            uploadPercentage={this.props.addAssetFilesModalUploadPercentage}

            onUploadCanceled={
              typeof this.props.addAssetFilesModalUploadPercentage === 'number'
              ? () => {
                if (currentFileUploadCancelToken === null) return
                Logger.debug('exerciseEditorSite', 'canceled file upload')
                currentFileUploadCancelToken.cancel('canceled')
              }
              : null
            }


            filterText={this.props.addAssetFilesModalFilterText}
            onFilterTextChanged={(text) => {
              this.props.setAddAssetFilesModalFilterText(text)
            }}

            attachedAssetFilePreviews={this.props.editorExercise.exerciseDescription.assets.map<FilePreviewFromBackend>(
              p => {
                return {
                  id: p.id,
                  originalName: p.displayName,
                  sizeInBytes: p.content.length,
                  mimeType: p.mimeType
                }
              })}

            allAssetFilePreviews={this.props.allMarkdownAssetFilePreviews}

            onAttachAssetFilePreview={async (assetFilePreview) => {
              this.props.setAddAssetFilesModalUploadPercentage('active')

              //this will download the file so display the progress bar
              try {
                await this.props.attachExitingFileToExerciseDescriptionAsync(this.props.editorExercise.id,
                                                                             assetFilePreview.id,
                                                                             // this.props.setAddAssetFilesModalUploadPercentage
                )
              } catch (err) {
                //not handled here
              } finally {
                this.props.setAddAssetFilesModalUploadPercentage(null)
              }

            }}

            onDetachAssetFilePreview={async (assetFilePreview) => {

              await this.props.detachFileFromExerciseDescriptionAsync(this.props.editorExercise.id, assetFilePreview.id)

              //refresh list
              await this.props.getMarkdownAssetPreviewsAsync(this.props.editorExercise.id)

            }}

            onClose={() => {
              this.props.setIsAddAssetFilesModalDisplayed(false, false)
            }}

          />
        }

        {
          this.props.isChangeTestDialogDisplayed &&
          <ChangeTestDialog/>
        }
        {
          this.props.isChangeSelectedTagsDialogDisplayed &&
          <ChangeSelectedTagsDialog/>
        }
        {
          this.props.isChangeCodeTemplateDialogDisplayed &&
          <ChangeCodeTemplateDialog/>
        }
        {
          this.props.isChangeEditorSettingsDialogDisplayed &&
          <EditorSettingsDialog/>
        }

        {
          this.props.isNoDialogDisplayed &&
          <PanelGrid
            width="100%"
            height="100%"

            topRightPanelViewSize={this.props.topRightPanelViewSize}
            topLeftPanelViewSize={this.props.topLeftPanelViewSize}
            botRightPanelViewSize={this.props.botRightPanelViewSize}
            botLeftPanelViewSize={this.props.botLeftPanelViewSize}

            hDividerPosOffset={this.props.hDividerPosOffset}
            vDividerPosOffset={this.props.vDividerPosOffset}
            onHDividerPosOffsetChanged={(Offset: number) => {
              this.props.setExerciseEditorHDividerPosOffset(Offset)
            }}
            onVDividerPosOffsetChanged={(Offset: number) => {
              this.props.setExerciseEditorVDividerPosOffset(Offset)
            }}

            topLeftPanel={
              <Panel
                applyDefaultPadding={false}
                selectedTabIndex={this.props.topLeftSelectedTabIndex}
                onSelectedTabChanged={(index) => {
                  this.props.setTopLeftSelectedTabIndex(index)
                }}
                onSizeChangedClicked={this.props.setTopLeftPanelViewSize}
                panelViewSize={this.props.topLeftPanelViewSize}
              >
                <TabContentHost
                  icon={'font'}
                  title={getI18n(this.props.langId, 'Description editor')}
                  applyDefaultPadding={false}
                  tabHeaderBar={
                    <DescriptionEditorHeaderBar
                      setIsEditorSettingsDialogDisplayed={this.props.setIsChangeExerciseEditorSettingsDialogDisplayed}
                    />
                  }
                >
                  <DescriptionEditorPanelView
                    descriptionContent={this.props.exerciseDescriptionContent}
                    setContent={(content: string) => {
                      this.props.setDescriptionContent(content)
                    }}/>
                </TabContentHost>

              </Panel>
            }

            topRightPanel={
              <Panel
                selectedTabIndex={this.props.topRightSelectedTabIndex}
                onSelectedTabChanged={(index) => {
                  this.props.setTopRightSelectedTabIndex(index)
                }}
                onSizeChangedClicked={this.props.setTopRightPanelViewSize}
                panelViewSize={this.props.topRightPanelViewSize}
                additionalActionsRightIcons={
                  this.props.topRightSelectedTabIndex === 1
                  ? <ZoomPreviewPanelActions printId={exerciseEditorTaskPrintId} canPrint={true}/>
                  : undefined
                }
              >

                <TabContentHost
                  title={getI18n(this.props.langId, 'Properties')}
                  icon={'options'}
                >
                  <ExerciseSettingsPanelView/>
                </TabContentHost>

                <TabContentHost
                  title={getI18n(this.props.langId, 'Preview')}
                  icon={'unhide'}
                >
                  <PreviewPanelView printId={exerciseEditorTaskPrintId}
                                    descriptionContent={this.props.editorExercise.exerciseDescription.content}
                                    assets={this.props.editorExercise.exerciseDescription.assets}
                                    dummy={false /*not needed because we have tabs that will re render if tab index changed*/}
                                    fontSizeInPx={this.props.taskDescriptionFontSizeInPx}
                  />
                </TabContentHost>

                <TabContentHost
                  title={getI18n(this.props.langId, 'Guide')}
                  icon={'book'}
                  tabHeaderBar={
                    <ExerciseSyntaxHelpPanelHeaderBar/>
                  }
                >
                  <ExerciseSyntaxHelpPanelView/>
                </TabContentHost>


              </Panel>
            }

            bottomLeftPanel={
              <Panel
                selectedTabIndex={this.props.bottomLeftSelectedTabIndex}
                onSelectedTabChanged={(index) => {
                  this.props.setBottomLeftSelectedTabIndex(index)
                }}
                onSizeChangedClicked={this.props.setBotLeftPanelViewSize}
                panelViewSize={this.props.botLeftPanelViewSize}
              >

                <TabContentHost
                  title={getI18n(this.props.langId, 'Code templates')}
                  icon={'file code outline'}
                  tabHeaderBar={
                    <TemplatesPanelHeaderBar/>
                  }
                >
                  <TemplatesPanelView/>
                </TabContentHost>

                <TabContentHost
                  title={getI18n(this.props.langId, 'Assets')}
                  icon={'file'}
                  tabHeaderBar={
                    <AssetsPanelHeaderBar
                      isEnabled={this.props.editorExercise.id > 0}
                      assets={this.props.assets}
                      setAssets={(assets) => {
                        this.props.setExerciseEditorAssets(assets)
                      }}
                      openAddAssetsModal={() => {
                        this.props.setIsAddAssetFilesModalDisplayed(true, false) //before loading because this will cancel isBusy
                        this.props.getMarkdownAssetPreviewsAsync(this.props.editorExercise.id)
                      }}
                    />
                  }
                >
                  <AssetsPanelView assets={this.props.assets}/>
                </TabContentHost>

                <TabContentHost
                  title={getI18n(this.props.langId, 'Tags')}
                  icon={'tags'}
                  tabHeaderBar={
                    <TagsPanelHeaderBar/>
                  }
                >
                  <TagsPanelView/>
                </TabContentHost>

                <TabContentHost
                  title={getI18n(this.props.langId, 'Note')}
                  icon={'sticky note'}
                >
                  <NotePanelView/>
                </TabContentHost>

              </Panel>
            }

            bottomRightPanel={
              <Panel
                selectedTabIndex={this.props.bottomRightSelectedTabIndex}
                onSelectedTabChanged={(index) => {
                  this.props.setBottomRightSelectedTabIndex(index)
                }}
                onSizeChangedClicked={this.props.setBotRightPanelViewSize}
                panelViewSize={this.props.botRightPanelViewSize}
              >
                <TabContentHost
                  title={getI18n(this.props.langId, 'Normal tests')}
                  icon={'bug'}
                  tabHeaderBar={
                    <ChangeTestsPanelHeaderBar isForSubmitTests={false}
                    />
                  }
                >
                  <ChangeTestsPanelView isForSubmitTests={false}
                  />
                </TabContentHost>

                <TabContentHost
                  icon="flag checkered"
                  title={getI18n(this.props.langId, 'Submit tests')}
                  tabHeaderBar={
                    <ChangeTestsPanelHeaderBar isForSubmitTests={true}
                    />
                  }
                >
                  <ChangeTestsPanelView isForSubmitTests={true}
                  />
                </TabContentHost>

              </Panel>
            }


          />
        }


      </div>
    )
  }
}

export default withRouter(connect(mapStateToProps, mapDispatchToProps)(exerciseEditorSite))
