/**
 * Created by janis dähne.
 */
import {
  ResetAction, SET_isLoadingManageGroupExercisesSiteACtion,
  SET_isTagsFilterDisplayedAction,
  SET_sortByKeyAction,
  SetActivePaginationPageAction, SetPaginationPageSizeAction, SetSearchTextAction
} from "../../reducers/manageGroupExercisesSite/groupExercisesSiteReducer";
import {EditableExercisePreviewFromBackend} from "../../../types/exercisePreview";
import {ActionType} from "../../reducers/manageGroupExercisesSite/groupExercisesSiteActionTypes";
import {AwaitActions, MultiActions} from "../types";
import {loadManageTagsSite} from "../manageTagsSite/manageTagsActions";
import {resetTagsFilter} from "../tagsFilterActions";
import {getGroupExercisesAsync, getGroupsForExercisesAsync} from './groupExercisesCrudSiteActions'
import debounce from 'lodash-es/debounce'
import {searchInputDebounceInMs} from '../../../constants'


export function setIsLoading(isLoading: boolean): SET_isLoadingManageGroupExercisesSiteACtion {
  return {
    type: ActionType.SET_isLoadingManageGroupExercisesSite,
    isLoading
  }
}




export function setIsTagsFilterDisplayed(isTagsFilterDisplayed: boolean): SET_isTagsFilterDisplayedAction {
  return {
    type: ActionType.SET_isTagsFilterDisplayed,
    isTagsFilterDisplayed,
  }
}

export function loadManageGroupExercisesSite(): AwaitActions {
  return async (dispatch, getState) => {

    dispatch(resetManageGroupExercisesSite())

    // dispatch(resetTagsFilter()) //in most cases we go back from release and then we want to still have the filter

    await dispatch(loadManageTagsSite())

    //load the groups
    await dispatch(getGroupsForExercisesAsync())

    //the load for every group the exercises (paginated)
    const groups = getState().manageGroupExercisesState.groups

    dispatch(setIsLoading(true))

    // noinspection TsLint
    for (let i = 0; i < groups.length; i++) {
      const group = groups[i]
      await dispatch(getGroupExercisesAsync(group.id)) //these only show loading for each group... but flickering...
    }
    dispatch(setIsLoading(false))

  }
}

export function resetManageGroupExercisesSite(): ResetAction {
  return {
    type: ActionType.RESET
  }
}



//--- pagination


export function _setSortByKey(sortByKey: keyof EditableExercisePreviewFromBackend,
                              sortDirection: boolean | null,
                              groupId: number
): SET_sortByKeyAction {
  return {
    type: ActionType.SET_sortByKey,
    sortByKey,
    sortDirection,
    groupId
  }
}

export function setSortByKey(sortByKey: keyof EditableExercisePreviewFromBackend,
                             sortDirection: boolean | null,
                             groupId: number
): MultiActions {
  return (dispatch, getState) => {

    dispatch(_setSortByKey(sortByKey, sortDirection, groupId))
    dispatch(getGroupExercisesAsync(groupId))

  }
}

export function _setActivePaginationPage(activePage: number, groupId: number): SetActivePaginationPageAction {
  return {
    type: ActionType.setPaginationActivePage,
    activePage,
    groupId
  }
}

export function setActivePaginationPage(activePage: number, groupId: number): MultiActions {
  return (dispatch, getState) => {

    dispatch(_setActivePaginationPage(activePage, groupId))
    dispatch(getGroupExercisesAsync(groupId))
  }
}

export function _setPaginationPageSize(pageSize: number, groupId: number): SetPaginationPageSizeAction {
  return {
    type: ActionType.setPaginationPageSize,
    pageSize,
    groupId
  }
}

export function setPaginationPageSize(pageSize: number, groupId: number): MultiActions {
  return (dispatch, getState) => {

    dispatch(_setPaginationPageSize(pageSize, groupId))

    dispatch(getGroupExercisesAsync(groupId))
  }
}

export function _setSearchText(searchText: string, groupId: number): SetSearchTextAction {
  return {
    type: ActionType.setSearchText,
    searchText,
    groupId
  }
}

// noinspection TsLint
const debouncedRefresh = debounce((dispatch, getState, groupId: number) => {

  dispatch(getGroupExercisesAsync(groupId))

}, searchInputDebounceInMs)

export function setSearchText(searchText: string, groupId: number): MultiActions {
  return (dispatch, getState) => {
    dispatch(_setSearchText(searchText, groupId))

    debouncedRefresh(dispatch, getState, groupId)
  }
}